@@ -1,16 +1,24 @@
 # -*- coding: utf-8 -*-
 # ------------------------------------------------------------------------------
 #
-#  HEBI Core python API - Copyright 2018 HEBI Robotics
+#  HEBI Core python API - Copyright 2022 HEBI Robotics
 #  See https://hebi.us/softwarelicense for license details
 #
 # ------------------------------------------------------------------------------
 
+# NOTE: This line _must_ be the first import!
 from ._internal.ffi import loader as __loader
+from . import arm, robot_model, trajectory, util
+from ._internal.graphics import Color
+from ._internal.ffi._message_types import GroupCommand, GroupFeedback, GroupInfo
 
 from . import version
 __version__ = str(version.py_version())
 
+import typing
+if typing.TYPE_CHECKING:
+  from typing import List, Optional, Union
+
 
 ################################################################################
 # Lookup API
@@ -21,10 +29,8 @@ from ._internal import lookup as _lookup_internal
 
 
 class Lookup(object):
-  """
-  Maintains a registry of network-connected modules
-  and returns :class:`hebi._internal.group.Group` objects to the user.
-  """
+  """Maintains a registry of network-connected modules and returns
+  :class:`hebi._internal.group.Group` objects to the user."""
 
   DEFAULT_TIMEOUT_MS = 500
   """
@@ -41,7 +47,7 @@ class Lookup(object):
     # We can't get lookup addresses from C API yet, so nothing to print re: that
     ret = 'lookup_frequency: {}\n\n'.format(lookup_freq)
     # Mobules table
-    from ._internal.utils import lookup_table_string  
+    from ._internal.utils import lookup_table_string
     return ret + lookup_table_string([entry for entry in self.entrylist])
 
   def __str__(self):
@@ -59,8 +65,7 @@ class Lookup(object):
 
   @property
   def entrylist(self):
-    """
-    A list of discovered network connected modules.
+    """A list of discovered network connected modules.
 
     :return: The list of modules
     :rtype: EntryList
@@ -75,9 +80,8 @@ class Lookup(object):
   def lookup_frequency(self, freq):
     self.__delegate.lookup_frequency = freq
 
-  def get_group_from_names(self, families, names, timeout_ms=None):
-    """
-    Get a group from modules with the given names and families.
+  def get_group_from_names(self, families: "Union[str, List[str]]", names: "Union[str, List[str]]", timeout_ms: "Optional[float]"=None):
+    """Get a group from modules with the given names and families.
 
     If the families or names provided as input is only a single element,
     then that element is assumed to pair with each item in the other parameter.
@@ -86,11 +90,11 @@ class Lookup(object):
     This will time out after :attr:`Lookup.DEFAULT_TIMEOUT_MS` milliseconds,
     if a matching group cannot be constructed.
 
-    :param families:   A list of families corresponding to the devices to include in the Group 
-    :type families:    list
+    :param families:   A family or list of families corresponding to the device(s) to include in the Group
+    :type families:    string or list
 
-    :param names:      A list of names corresponding to the devices to include in the Group
-    :type names:       list
+    :param names:      A name or list of names corresponding to the device(s) to include in the Group
+    :type names:       string or list
 
     :param timeout_ms: The maximum amount of time to wait, in milliseconds.
                        This is an optional parameter with a default value of
@@ -99,11 +103,14 @@ class Lookup(object):
     :return: A group on success; ``None`` if one or more devices specified could not be found
     :rtype: hebi._internal.group.Group
     """
+    if not isinstance(families, list):
+      families = [families]
+    if not isinstance(names, list):
+      names = [names]
     return self.__delegate.get_group_from_names(families, names, timeout_ms)
 
   def get_group_from_macs(self, addresses, timeout_ms=None):
-    """
-    Get a group from modules with the given mac addresses.
+    """Get a group from modules with the given mac addresses.
 
     This is a blocking call which returns a Group with the given parameters.
     This will time out after :attr:`Lookup.DEFAULT_TIMEOUT_MS` milliseconds,
@@ -127,9 +134,8 @@ class Lookup(object):
     """
     return self.__delegate.get_group_from_macs(addresses, timeout_ms)
 
-  def get_group_from_family(self, family, timeout_ms=None):
-    """
-    Get a group from all known modules with the given family.
+  def get_group_from_family(self, family: str, timeout_ms=None):
+    """Get a group from all known modules with the given family.
 
     This is a blocking call which returns a Group with the given parameters.
     This will time out after :attr:`Lookup.DEFAULT_TIMEOUT_MS` milliseconds,
@@ -148,9 +154,8 @@ class Lookup(object):
     return self.__delegate.get_group_from_family(family, timeout_ms)
 
   def get_connected_group_from_name(self, family, name, timeout_ms=None):
-    """
-    Get a group from all modules known to connect to a module with
-    the given name and family.
+    """Get a group from all modules known to connect to a module with the given
+    name and family.
 
     This is a blocking call which returns a Group with the given parameters.
     This will time out after :attr:`Lookup.DEFAULT_TIMEOUT_MS` milliseconds,
@@ -161,7 +166,7 @@ class Lookup(object):
 
     :param name:       The name of the connected device, used to create the Group
     :type name:        str
-    
+
     :param timeout_ms: The maximum amount of time to wait, in milliseconds.
                        This is an optional parameter with a default value of
                        :attr:`Lookup.DEFAULT_TIMEOUT_MS`.
@@ -173,9 +178,8 @@ class Lookup(object):
     return self.__delegate.get_connected_group_from_name(family, name, timeout_ms)
 
   def get_connected_group_from_mac(self, address, timeout_ms=None):
-    """
-    Get a group from all modules known to connect to a module with
-    the given mac address.
+    """Get a group from all modules known to connect to a module with the given
+    mac address.
 
     This is a blocking call which returns a Group with the given parameters.
     This will time out after :attr:`Lookup.DEFAULT_TIMEOUT_MS` milliseconds,
@@ -206,9 +210,5 @@ class Lookup(object):
 ################################################################################
 
 
-from ._internal.ffi._message_types import GroupCommand, GroupFeedback, GroupInfo
-from ._internal.graphics import Color
-from . import arm, robot_model, trajectory, util
-
 __all__ = ['Lookup', 'GroupCommand', 'GroupFeedback', 'GroupInfo', 'Color',
            'robot_model', 'trajectory', 'util', 'arm']
