@@ -26,7 +26,7 @@ from numpy.ctypeslib import as_array as _ctypes_to_ndarray
 
 import typing
 if typing.TYPE_CHECKING:
-  from typing import Any, Callable, List, Optional, TypeVar, Union, Iterable
+  from typing import Any, Callable, List, Optional, TypeVar, Union, Iterable, Sequence
   import numpy.typing as npt
   from ._message_types import Command, Info, GroupCommandBase, GroupFeedbackBase, GroupInfoBase
   HebiBase = Union[GroupCommandBase, GroupFeedbackBase, GroupInfoBase]
@@ -1262,19 +1262,19 @@ def __set_bool_group(refs: "Array[HebiCommandRef]", field: MessageEnumTraits, va
   setter(refs, bfr, size, field)
 
 
-def __set_enum_group(refs: "Array[HebiCommandRef]", field: MessageEnumTraits, value, setter):
+def __set_enum_group(refs: "Array[HebiCommandRef]", field: MessageEnumTraits, value: "Optional[Union[Sequence[int], int]]", setter):
   size = len(refs)
   _tls.ensure_capacity(size)
   if value is None:
     bfr = None
   else:
     bfr = _tls.c_int32_array
-    if hasattr(value, '__len__'):
-      for i in range(size):
-        bfr[i] = value[i]
-    else:
+    if isinstance(value, int):
       for i in range(size):
         bfr[i] = value
+    else:
+      for i in range(size):
+        bfr[i] = value[i]
   setter(refs, bfr, size, field)
 
 
@@ -1477,7 +1477,7 @@ def set_group_command_flag(msg: "Array[HebiCommandRef]", field: MessageEnumTrait
 def set_group_command_bool(msg: "Array[HebiCommandRef]", field: MessageEnumTraits, value: "Union[npt.NDArray[np.bool_], bool, None]"):
   __set_bool_group(msg, field, value, api.hwCommandSetBool)
 
-def set_group_command_enum(msg: "Array[HebiCommandRef]", field: MessageEnumTraits, value: "List[int]"):
+def set_group_command_enum(msg: "Array[HebiCommandRef]", field: MessageEnumTraits, value: "Union[Sequence[int], int, None]"):
   __set_enum_group(msg, field, value, api.hwCommandSetEnum)
 
 def set_group_command_float(msg: "Array[HebiCommandRef]", field: MessageEnumTraits, value: "Union[npt.NDArray[np.float32], float, None]"):
@@ -1490,7 +1490,7 @@ def set_group_command_string(msg: "HebiBase", field: MessageEnumTraits, value: "
   alloc_size_c = _tls.c_size_t
   if value is not None:
     for message in msg.modules:
-      alloc_size = len(value) + 1
+      alloc_size = len(value.encode('utf-8')) + 1
       # TODO: use tls string buffer and copy val into it instead
       string_buffer = create_str(value, size=alloc_size)
       alloc_size_c.value = alloc_size
