@@ -5,8 +5,9 @@ from .ctypes_utils import cast, c_int_p, c_int32_p, c_uint32_p, c_int64_p, c_uin
 
 
 class TLS_Holder:
-  __slots__ = ["_buffer", "_size",
-              "_c_int_p", "_c_int32_p", "_c_uint32_p", "_c_int64_p", "_c_uint64_p", "_c_float_p", "_c_double_p"
+  __slots__ = [
+      "_buffer", "_size", "_c_int_p", "_c_int32_p", "_c_uint32_p",
+      "_c_int64_p", "_c_uint64_p", "_c_float_p", "_c_double_p"
   ]
 
   def _grow_arrays(self, size):
