@@ -1,3 +1,5 @@
+from ctypes import CDLL
+
 _hebi_api = None
 _hebi_wrapper_api = None
 
@@ -59,11 +61,14 @@ hebiGroupInfoClear = None
 hebiGroupInfoRelease = None
 hebiCommandGetString = None
 hebiCommandSetString = None
+hebiCommandGetIoLabelString = None
+hebiCommandSetIoLabelString = None
 hebiCommandGetReference = None
 hebiCommandGetMetadata = None
 hebiFeedbackGetReference = None
 hebiFeedbackGetMetadata = None
 hebiInfoGetString = None
+hebiInfoGetIoLabelString = None
 hebiInfoGetReference = None
 hebiInfoGetMetadata = None
 hebiRobotModelElementCreateJoint = None
@@ -167,13 +172,10 @@ hwInfoGetLed = None
 hwInfoHasField = None
 
 
-def set_api_handle(lib):
-  """
-  Sets the handle to the HEBI Core C API
-  """
+def set_api_handle(lib: CDLL):
+  """Sets the handle to the HEBI Core C API."""
   global _hebi_api
   _hebi_api = lib
-  
 
   global hebiLookupCreate
   hebiLookupCreate = lib.hebiLookupCreate
@@ -289,6 +291,10 @@ def set_api_handle(lib):
   hebiCommandGetString = lib.hebiCommandGetString
   global hebiCommandSetString
   hebiCommandSetString = lib.hebiCommandSetString
+  global hebiCommandGetIoLabelString
+  hebiCommandGetIoLabelString = lib.hebiCommandGetIoLabelString
+  global hebiCommandSetIoLabelString
+  hebiCommandSetIoLabelString = lib.hebiCommandSetIoLabelString
   global hebiCommandGetReference
   hebiCommandGetReference = lib.hebiCommandGetReference
   global hebiCommandGetMetadata
@@ -299,6 +305,8 @@ def set_api_handle(lib):
   hebiFeedbackGetMetadata = lib.hebiFeedbackGetMetadata
   global hebiInfoGetString
   hebiInfoGetString = lib.hebiInfoGetString
+  global hebiInfoGetIoLabelString
+  hebiInfoGetIoLabelString = lib.hebiInfoGetIoLabelString
   global hebiInfoGetReference
   hebiInfoGetReference = lib.hebiInfoGetReference
   global hebiInfoGetMetadata
@@ -401,13 +409,10 @@ def set_api_handle(lib):
   hebiCleanup = lib.hebiCleanup
 
 
-def set_api_wrapper_handle(lib):
-  """
-  Sets the handle to the C wrapper API
-  """
+def set_api_wrapper_handle(lib: CDLL):
+  """Sets the handle to the C wrapper API."""
   global _hebi_wrapper_api
   _hebi_wrapper_api = lib
-  
 
   global hwInitialize
   hwInitialize = lib.hwInitialize
