@@ -195,7 +195,7 @@ else:
 # CTypes Compatibility functions
 ################################################################################
 
-def create_string_buffer_compat(init, size=None):
+def create_string_buffer_compat(init, size: "Optional[int]"=None):
   if size is None:
     if isinstance(init, str):
       return create_string_buffer(bytes(init, 'utf8'))
