@@ -1,7 +1,7 @@
 # -*- coding: utf-8 -*-
 # ------------------------------------------------------------------------------
 #
-#  HEBI Core python API - Copyright 2018 HEBI Robotics
+#  HEBI Core python API - Copyright 2022 HEBI Robotics
 #  See https://hebi.us/softwarelicense for license details
 #
 # ------------------------------------------------------------------------------
@@ -16,13 +16,17 @@ from ._internal.ffi.ctypes_utils import byref, c_double_p, to_double_ptr, NULLPT
 
 from ._internal.ffi.wrappers import UnmanagedObject
 from ._internal.ffi import enums
-from ._internal.ffi import ctypes_func_defs as api 
+from ._internal.ffi import ctypes_func_defs as api
 
-import ctypes as _ctypes
-import math as _math
 import numpy as _np
 import threading as _threading
 
+import typing
+if typing.TYPE_CHECKING:
+  from typing import Union, Optional, Sequence, Callable, Any, List
+  import numpy.typing as npt
+  VectorType = Union[Sequence[float], npt.NDArray[_np.float64]]
+
 
 class _RobotModelTLS_Buffers(object):
 
@@ -30,10 +34,10 @@ class _RobotModelTLS_Buffers(object):
 
   def __init__(self):
     self._base_frame = _type_utils.create_double_buffer(16)
-    self._inertia = _np.empty(6, dtype=_np.float64)
+    self._inertia: "npt.NDArray[_np.float64]" = _np.empty(6, dtype=_np.float64)
     self._double_buffer = _type_utils.create_double_buffer(256)
-    self._com_transform = _np.identity(4, dtype=_np.float64)
-    self._output_transform = _np.identity(4, dtype=_np.float64)
+    self._com_transform: "npt.NDArray[_np.float64]" = _np.identity(4, dtype=_np.float64)
+    self._output_transform: "npt.NDArray[_np.float64]" = _np.identity(4, dtype=_np.float64)
 
   @property
   def base_frame(self):
@@ -61,12 +65,12 @@ class _RobotModelTLS_Buffers(object):
 
 
 class _RobotModelTLS(_threading.local):
-  """
-  Thread local buffers for the RobotModel API.
+  """Thread local buffers for the RobotModel API.
 
   Note: this class **cannot** have the `__slots__` attribute, as it breaks
         the `threading.local` API.
   """
+
   def __init__(self):
     self._buffers = _RobotModelTLS_Buffers()
 
@@ -99,8 +103,8 @@ _row_major = enums.MatrixOrderingRowMajor
 
 
 class RobotModel(UnmanagedObject):
-  """
-  Represents a chain or tree of robot elements (rigid bodies and joints).
+  """Represents a chain or tree of robot elements (rigid bodies and joints).
+
   Currently, only chains of elements are fully supported.
   """
   __slots__ = ['_dof_count', '_element_count', '_frame_count_map', '_masses_cached']
@@ -117,11 +121,11 @@ class RobotModel(UnmanagedObject):
 
     self._dof_count = 0
     self._element_count = 0
-    self._frame_count_map = dict()
+    self._frame_count_map: dict[enums.EnumTraits, int] = dict()
     self._frame_count_map[enums.FrameTypeCenterOfMass] = 0
     self._frame_count_map[enums.FrameTypeOutput] = 0
     self._frame_count_map[enums.FrameTypeEndEffector] = 0
-    self._masses_cached = _np.empty(0, _np.float64)
+    self._masses_cached: npt.NDArray[_np.float64] = _np.empty(0, _np.float64)
 
     if has_existing_handle:
       self.__update_cached_frame_counts()
@@ -135,16 +139,14 @@ class RobotModel(UnmanagedObject):
 
     # Note: The C API does not support payloads yet. This is intentionally omitted.
     #payload = self.payload
-    ret = 'RobotModel with properties:\n'                       +\
-          '\n'                                                  +\
-          '    body_count: {0}\n'.format(body_count)            +\
-          '     dof_count: {0}\n'.format(dof_count)             +\
-          '          mass: {:0.2f} [kg]\n'.format(total_mass)
-
-    ret = ret +\
-          '\n'                                                  +\
-          '    body  type             is_dof  mass [kg]\n'      +\
-          '    ----  ---------------  ------  ---------\n'
+    ret = ('RobotModel with properties:\n' +
+           '\n' +
+           '    body_count: {0}\n'.format(body_count) +
+           '     dof_count: {0}\n'.format(dof_count) +
+           '          mass: {:0.2f} [kg]\n'.format(total_mass) +
+           '\n' +
+           '    body  type             is_dof  mass [kg]\n' +
+           '    ----  ---------------  ------  ---------\n')
 
     for i in range(body_count):
       elem_metadata = metadata[i]
@@ -159,8 +161,8 @@ class RobotModel(UnmanagedObject):
 
   @property
   def base_frame(self):
-    """
-    The transform from the world coordinate system to the root kinematic body
+    """The transform from the world coordinate system to the root kinematic
+    body.
 
     :return: The base frame 4x4 matrix
     :rtype:  numpy.ndarray
@@ -174,9 +176,8 @@ class RobotModel(UnmanagedObject):
 
   @base_frame.setter
   def base_frame(self, value):
-    """
-    Set the transform from a world coordinate system to the input of the root
-    element in this robot model. Defaults to an identity 4x4 matrix.
+    """Set the transform from a world coordinate system to the input of the
+    root element in this robot model. Defaults to an identity 4x4 matrix.
 
     The world coordinate system is used for all position, vector,
     and transformation matrix parameters in the member functions.
@@ -200,8 +201,8 @@ class RobotModel(UnmanagedObject):
 
   @property
   def metadata(self):
-    """
-    Retrieves a list of info about each individual element which composes this robot
+    """Retrieves a list of info about each individual element which composes
+    this robot.
 
     :rtype: list
     """
@@ -216,8 +217,8 @@ class RobotModel(UnmanagedObject):
 
   @property
   def element_count(self):
-    """
-    The number of elements which compose the kinematic tree. This is greater than or equal to the degrees of freedom.
+    """The number of elements which compose the kinematic tree. This is greater
+    than or equal to the degrees of freedom.
 
     :rtype: int
     """
@@ -226,9 +227,8 @@ class RobotModel(UnmanagedObject):
   def __get_frame_count(self, frame_type):
     return self._frame_count_map[frame_type]
 
-  def get_frame_count(self, frame_type):
-    """
-    The number of frames in the forward kinematics.
+  def get_frame_count(self, frame_type: str):
+    """The number of frames in the forward kinematics.
 
     Note that this depends on the type of frame requested:
       * for center of mass frames, there is one per added body.
@@ -252,9 +252,8 @@ class RobotModel(UnmanagedObject):
 
   @property
   def dof_count(self):
-    """
-    The number of settable degrees of freedom in the kinematic tree.
-    This is equal to the number of actuators added.
+    """The number of settable degrees of freedom in the kinematic tree. This is
+    equal to the number of actuators added.
 
     :return: the degrees of freedom.
     :rtype:  int
@@ -262,7 +261,7 @@ class RobotModel(UnmanagedObject):
     return self._dof_count
 
   def __update_cached_masses(self, num_com_frames):
-    masses = _np.empty(num_com_frames, dtype=_np.float64)
+    masses: npt.NDArray[_np.float64] = _np.empty(num_com_frames, dtype=_np.float64)
     api.hebiRobotModelGetMasses(self, to_double_ptr(masses))
     self._masses_cached = masses
 
@@ -289,8 +288,7 @@ class RobotModel(UnmanagedObject):
     return True
 
   def add_rigid_body(self, com, inertia, mass, output):
-    """
-    Adds a rigid body with the specified properties to the robot model.
+    """Adds a rigid body with the specified properties to the robot model.
 
     This can be 'combined' with the parent element
     (the element to which this is attaching), which means that the mass,
@@ -340,13 +338,12 @@ class RobotModel(UnmanagedObject):
     elif user_com.shape == (4, 4):
       _np.copyto(com, user_com)
 
-    body = api.hebiRobotModelElementCreateRigidBody(to_double_ptr(com), to_double_ptr(inertia), mass, 1, to_double_ptr(output), _get_matrix_ordering(output))
+    body = api.hebiRobotModelElementCreateRigidBody(to_double_ptr(com), to_double_ptr(inertia), mass,
+                                                    1, to_double_ptr(output), _get_matrix_ordering(output))
     return self.__try_add(body)
 
-
   def add_joint(self, joint_type):
-    """
-    Adds a degree of freedom about the specified axis.
+    """Adds a degree of freedom about the specified axis.
 
     This does not represent an element with size or mass, but only a
     connection between two other elements about a particular axis.
@@ -368,10 +365,9 @@ class RobotModel(UnmanagedObject):
     """
     return self.__try_add(api.hebiRobotModelElementCreateJoint(_kinematics.parse_joint_type(joint_type)))
 
-  def add_actuator(self, actuator_type):
-    """
-    Add an element to the robot model with the kinematics/dynamics
-    of an X or R series HEBI actuator.
+  def add_actuator(self, actuator_type: str):
+    """Add an element to the robot model with the kinematics/dynamics of an X
+    or R series HEBI actuator.
 
     :param actuator_type: The type of actuator to add.
     :type actuator_type:  str, unicode
@@ -384,10 +380,9 @@ class RobotModel(UnmanagedObject):
     """
     return self.__try_add(api.hebiRobotModelElementCreateActuator(_kinematics.actuator_str_to_enum(actuator_type)))
 
-  def add_link(self, link_type, extension, twist):
-    """
-    Add an element to the robot model with the kinematics/dynamics
-    of a link between two actuators.
+  def add_link(self, link_type: str, extension: float, twist: float):
+    """Add an element to the robot model with the kinematics/dynamics of a link
+    between two actuators.
 
     :param link_type: The type of link between the actuators, e.g. a tube link
                       between two X5 or X8 actuators.
@@ -416,13 +411,12 @@ class RobotModel(UnmanagedObject):
 
     input_type = enums.LinkInputTypeRightAngle
     output_type = enums.LinkOutputTypeRightAngle
-    link_type = _kinematics.link_str_to_enum(link_type)
-    return self.__try_add(api.hebiRobotModelElementCreateLink(link_type, input_type, output_type, extension, twist))
+    link_enum = _kinematics.link_str_to_enum(link_type)
+    return self.__try_add(api.hebiRobotModelElementCreateLink(link_enum, input_type, output_type, extension, twist))
 
-  def add_bracket(self, bracket_type, mount):
-    """
-    Add an element to the robot model with the kinematics/dynamics of a bracket
-    between two actuators.
+  def add_bracket(self, bracket_type: str, mount: str):
+    """Add an element to the robot model with the kinematics/dynamics of a
+    bracket between two actuators.
 
     :param bracket_type: The type of bracket to add.
     :type bracket_type:  str, unicode
@@ -438,10 +432,9 @@ class RobotModel(UnmanagedObject):
     """
     return self.__try_add(api.hebiRobotModelElementCreateBracket(_kinematics.bracket_str_to_enum(bracket_type, mount)))
 
-  def add_end_effector(self, end_effector_type):
-    """
-    Add an end effector element to the robot model.
-   
+  def add_end_effector(self, end_effector_type: str):
+    """Add an end effector element to the robot model.
+
     For a "custom" type end effector, indentity transforms and
     zero mass and inertia parameters are used.
 
@@ -454,13 +447,12 @@ class RobotModel(UnmanagedObject):
     :raises ValueError: If the string from ``end_effector_type`` is invalid
     :raises TypeError:  If the ``end_effector_type`` argument is not a string
     """
-    end_effector_type = _kinematics.end_effector_str_to_enum(end_effector_type)
-    body = api.hebiRobotModelElementCreateEndEffector(end_effector_type, None, None, 0, None, _row_major)
+    end_effector_enum = _kinematics.end_effector_str_to_enum(end_effector_type)
+    body = api.hebiRobotModelElementCreateEndEffector(end_effector_enum, None, None, 0, None, _row_major)
     return self.__try_add(body)
 
-  def get_forward_kinematics(self, frame_type, positions, output=None):
-    """
-    Generates the forward kinematics for the given robot model.
+  def get_forward_kinematics(self, frame_type: str, positions: "VectorType", output: "Optional[List[npt.NDArray[_np.float64]]]"=None):
+    """Generates the forward kinematics for the given robot model.
 
     The order of the returned frames is in a depth-first tree.
 
@@ -493,8 +485,8 @@ class RobotModel(UnmanagedObject):
     :raises ValueError: If the ``positions`` input is not equal to the
                         degrees of freedom of the RobotModel
     """
-    frame_type = _kinematics.parse_frame_type(frame_type)
-    num_frames = self.__get_frame_count(frame_type)
+    frame_type_enum = _kinematics.parse_frame_type(frame_type)
+    num_frames = self.__get_frame_count(frame_type_enum)
 
     positions = _np.asarray(positions, _np.float64)
     self.__assert_equals_dof_count(positions)
@@ -508,23 +500,19 @@ class RobotModel(UnmanagedObject):
     _tls.buffers.grow_double_buffer_if_needed(num_frames * 16)
     frames = _tls.buffers.double_buffer
 
-    api.hebiRobotModelGetForwardKinematics(self, frame_type, to_double_ptr(positions), frames, _row_major)
+    api.hebiRobotModelGetForwardKinematics(self, frame_type_enum, to_double_ptr(positions), frames, _row_major)
 
     if output is None:
-      mat = _np.empty((4, 4), dtype=_np.float64)
-      output = [None] * num_frames
-      output[0] = mat
-      for i in range(1, num_frames):
-        output[i] = mat.copy()
-    for i in range(0, num_frames):
+      mat: npt.NDArray[_np.float64] = _np.empty((4, 4), dtype=_np.float64)
+      output = [mat.copy() for _ in range(num_frames)]
+    for i in range(num_frames):
       start = i * 16
       end = start + 16
       _np.copyto(output[i].ravel(), frames[start:end])
     return output
 
-  def get_end_effector(self, positions, output=None):
-    """
-    Generates the forward kinematics to the end effector (leaf node)
+  def get_end_effector(self, positions: "VectorType", output: "Optional[npt.NDArray[_np.float64]]"=None):
+    """Generates the forward kinematics to the end effector (leaf node)
 
     Note: for center of mass frames, this is one per leaf node; for output
     frames, this is one per output per leaf node, in depth first order.
@@ -572,9 +560,8 @@ class RobotModel(UnmanagedObject):
       _np.copyto(output.ravel(), transforms)
       return output
 
-  def solve_inverse_kinematics(self, initial_positions, *objectives, **kwargs):
-    """
-    Solves for an inverse kinematics solution given a set of objectives.
+  def solve_inverse_kinematics(self, initial_positions: "VectorType", *objectives: "_ObjectiveBase", **kwargs: "Any"):
+    """Solves for an inverse kinematics solution given a set of objectives.
 
     To avoid unnecessary allocations, provide ``output`` as a keyword argument.
     This argument must be a numpy array or matrix with dtype
@@ -637,25 +624,26 @@ class RobotModel(UnmanagedObject):
 
     if code != enums.StatusSuccess:
       raise HEBI_Exception(code, 'hebiIKSolve failed')
+
     if has_output:
-      output = kwargs['output']
-      _np.copyto(output.ravel(), result[0:dof_count])
-      return output
-    return _np.asarray(result[0:dof_count], _np.float64)
+      output: npt.NDArray[_np.float64] = kwargs['output']
+    else:
+      output = _np.empty(dof_count, dtype=_np.float64)
+
+    _np.copyto(output.ravel(), result[0:dof_count])
+    return output
+    #return _np.asarray(result[0:dof_count], dtype=_np.float64)
 
   def __get_jacobians(self, frame_type, positions, rows, cols):
-    """
-    Callee is required to copy results back from TLS buffer
-    """
+    """Callee is required to copy results back from TLS buffer."""
     buffers = _tls.buffers
     buffers.grow_double_buffer_if_needed(rows * cols)
     ret = buffers.double_buffer
     api.hebiRobotModelGetJacobians(self, frame_type, positions, ret, _row_major)
     return ret
 
-  def get_jacobians(self, frame_type, positions, output=None):
-    """
-    Generates the Jacobian for each frame in the given kinematic tree.
+  def get_jacobians(self, frame_type: str, positions: "VectorType", output: "Optional[List[npt.NDArray[_np.float64]]]"=None):
+    """Generates the Jacobian for each frame in the given kinematic tree.
 
     Deprecation notice: It is deprecated to pass a `str` in as a parameter for `positions`.
     This functionality will be removed in a future release.
@@ -682,8 +670,8 @@ class RobotModel(UnmanagedObject):
               on the robot. It is resized as necessary inside this function.
     :rtype:   list
     """
-    frame_type = _kinematics.parse_frame_type(frame_type)
-    num_frames = self.__get_frame_count(frame_type)
+    frame_type_enum = _kinematics.parse_frame_type(frame_type)
+    num_frames = self.__get_frame_count(frame_type_enum)
 
     # Edge case
     if num_frames == 0:
@@ -700,24 +688,20 @@ class RobotModel(UnmanagedObject):
 
     if output is None:
       mat = _np.empty((6, cols), dtype=_np.float64)
-      output = [None] * num_frames
-      output[0] = mat
-      for i in range(1, num_frames):
-        output[i] = mat.copy()
+      output = [mat.copy() for _ in range(num_frames)]
 
-    jacobians = self.__get_jacobians(frame_type, to_double_ptr(positions), rows, cols)
+    jacobians = self.__get_jacobians(frame_type_enum, to_double_ptr(positions), rows, cols)
 
     stride = cols * 6
     for i in range(0, num_frames):
       start = i * stride
       end = start + stride
       _np.copyto(output[i].ravel(), jacobians[start:end])
-    
+
     return output
 
   def get_jacobian_end_effector(self, positions, output=None):
-    """
-    Generates the Jacobian for the end effector (leaf node) frames(s).
+    """Generates the Jacobian for the end effector (leaf node) frames(s).
 
     Note: for center of mass frames, this is one per leaf node; for output
     frames, this is one per output per leaf node, in depth first order.
@@ -760,7 +744,7 @@ class RobotModel(UnmanagedObject):
 
     if output is None:
       output = _np.zeros((rows, cols), dtype=_np.float64)
-  
+
     jacobians = self.__get_jacobians(enums.FrameTypeEndEffector, to_double_ptr(positions), rows, cols)
 
     stride = cols * rows
@@ -771,14 +755,94 @@ class RobotModel(UnmanagedObject):
 
   @property
   def masses(self):
-    """
-    The mass of each rigid body (or combination of rigid bodies) in the robot.
+    """The mass of each rigid body (or combination of rigid bodies) in the
+    robot.
 
     :return: The masses as an array
     :rtype:  numpy.ndarray
     """
     return self._masses_cached.copy()
 
+  def get_grav_comp_efforts(self, positions, gravity, output=None):
+    """
+    :param positions:
+    :param gravity:
+
+    :param output:
+    :type output:  np.ndarray, NoneType
+
+    :return:
+    :rtype:  np.ndarray
+    """
+    g_norm = _np.linalg.norm(gravity)
+    if g_norm > 0.0:
+      gravity = gravity/g_norm*9.81
+
+    jacobians = self.get_jacobians('CoM', positions)
+    if output is None:
+      comp_torque = _np.zeros(self.dof_count)
+    else:
+      comp_torque = output
+    wrench = _np.zeros(6)
+    num_frames = self.get_frame_count('CoM')
+
+    masses = self.masses
+
+    for i in range(num_frames):
+      # Add the torques for each joint to support the mass at this frame
+      wrench[0:3] = gravity*masses[i]
+      comp_torque += jacobians[i].T @ wrench
+
+    return comp_torque
+
+  def get_dynamic_comp_efforts(self, fbk_positions, cmd_positions, cmd_velocities, cmd_accels, dt=1e-3):
+    """
+    :param fbk_positions:
+    :param cmd_positions:
+    :param cmd_velocities:
+    :param cmd_accels:
+    :param robot:
+    :param dt:
+
+    :return:
+    :rtype:  np.ndarray
+    """
+    dt_s = dt*dt
+    dt_s_inv = 1/dt_s
+    cmd_v_dt = cmd_velocities*dt
+    cmd_accel_dt = .5*cmd_accels*dt_s
+
+    # get positions at +/- dt
+    cmd_positions_last = cmd_positions-cmd_v_dt+cmd_accel_dt
+    cmd_positions_next = cmd_positions+cmd_v_dt+cmd_accel_dt
+
+    # Get Forward kinematics for all 3 sets of angles
+    cmd_frames_last = [entry[:3, 3] for entry in self.get_forward_kinematics('CoM', cmd_positions_last)]
+    cmd_frames_now = [entry[:3, 3] for entry in self.get_forward_kinematics('CoM', cmd_positions)]
+    cmd_frames_next = [entry[:3, 3] for entry in self.get_forward_kinematics('CoM', cmd_positions_next)]
+
+    # Build wrench vector and calculate compensatory torques
+    efforts: npt.NDArray[_np.float64] = _np.zeros(self.dof_count, dtype=_np.float64)
+    jacobians = self.get_jacobians('CoM', fbk_positions)
+    masses = self.masses
+    wrench: npt.NDArray[_np.float64] = _np.zeros(6, dtype=_np.float64)
+
+    for module in range(len(masses)):
+      # Calculate XYZ accelerations of the CoM
+      lastv = cmd_frames_last[module]
+      nowv = cmd_frames_now[module]
+      nextv = cmd_frames_next[module]
+
+      accel = ((lastv+nextv)-(2*nowv))*dt_s_inv
+
+      # Set translational part of wrench vector (rotational stays zero)
+      wrench[0:3] = accel * masses[module]
+
+      # compEffort = J' * wrench
+      efforts += jacobians[module].T @ wrench
+
+    return efforts
+
 
 ################################################################################
 # Robot Model metadata functionality
@@ -786,9 +850,7 @@ class RobotModel(UnmanagedObject):
 
 
 class OtherMetaData(object):
-  """
-  Metadata pertaining to an unknown object
-  """
+  """Metadata pertaining to an unknown object."""
 
   def __init__(self):
     pass
@@ -816,14 +878,12 @@ class OtherMetaData(object):
 
 
 class ActuatorMetaData(object):
-  """
-  Metadata pertaining to an actuator
-  """
+  """Metadata pertaining to an actuator."""
 
   __slots__ = ['_actuator_type']
 
   def __init__(self, actuator_type):
-      self._actuator_type = actuator_type
+    self._actuator_type = actuator_type
 
   def __repr__(self):
     return "Type:  Actuator\n" +\
@@ -838,7 +898,7 @@ class ActuatorMetaData(object):
     :return: The enum value corresponding to the specific actuator type of this element
     """
     return self._actuator_type
-  
+
   @property
   def type(self):
     """
@@ -856,9 +916,7 @@ class ActuatorMetaData(object):
 
 
 class BracketMetaData(object):
-  """
-  Metadata pertaining to a bracket
-  """
+  """Metadata pertaining to a bracket."""
 
   __slots__ = ['_bracket_type']
 
@@ -878,7 +936,7 @@ class BracketMetaData(object):
     :return: The enum value corresponding to the specific bracket type of this element
     """
     return self._bracket_type
-  
+
   @property
   def type(self):
     """
@@ -904,9 +962,7 @@ class BracketMetaData(object):
 
 
 class JointMetaData(object):
-  """
-  Metadata pertaining to a joint
-  """
+  """Metadata pertaining to a joint."""
 
   __slots__ = ['_joint_type']
 
@@ -926,7 +982,7 @@ class JointMetaData(object):
     :return: The enum value corresponding to the specific joint type of this element
     """
     return self._joint_type
-  
+
   @property
   def type(self):
     """
@@ -952,9 +1008,7 @@ class JointMetaData(object):
 
 
 class LinkMetaData(object):
-  """
-  Metadata pertaining to a link
-  """
+  """Metadata pertaining to a link."""
 
   __slots__ = ['_extension', '_link_type', '_twist']
 
@@ -1019,9 +1073,7 @@ class LinkMetaData(object):
 
 
 class RigidBodyMetaData(object):
-  """
-  Metadata pertaining to a rigid body
-  """
+  """Metadata pertaining to a rigid body."""
 
   __slots__ = []
 
@@ -1059,9 +1111,7 @@ class RigidBodyMetaData(object):
 
 
 class EndEffectorMetaData(object):
-  """
-  Metadata pertaining to an end effector
-  """
+  """Metadata pertaining to an end effector."""
 
   __slots__ = ['_end_effector_type']
 
@@ -1081,7 +1131,7 @@ class EndEffectorMetaData(object):
     :return: The enum value corresponding to the specific end effector type of this element
     """
     return self._end_effector_type
-  
+
   @property
   def type(self):
     """
@@ -1171,9 +1221,8 @@ def _get_all_hrdf_import_warnings():
   return ret
 
 
-def import_from_hrdf(hrdf_file, warning_callback=None):
-  """
-  Import a robot description in the HRDF format as a RobotModel instance.
+def import_from_hrdf(hrdf_file: str, warning_callback: "Optional[Callable[[Any], None]]"=None):
+  """Import a robot description in the HRDF format as a RobotModel instance.
 
   Any warnings generated while importing will be printed to stderr,
   unless explicitly instructed otherwise by providing a callback.
@@ -1217,9 +1266,9 @@ def import_from_hrdf(hrdf_file, warning_callback=None):
 
 
 def import_from_hrdf_string(hrdf_string, warning_callback=None):
-  """
-  Provides same functionality as :meth:`.import_from_hrdf`, but accepts a string representing the HRDF as opposed
-  to a string to the file representing the HRDF.
+  """Provides same functionality as :meth:`.import_from_hrdf`, but accepts a
+  string representing the HRDF as opposed to a string to the file representing
+  the HRDF.
 
   :param hrdf_string: A string representing HRDF contents
   :type hrdf_string:  str
@@ -1259,21 +1308,21 @@ class _ObjectiveBase(object):
     self.__impl(internal)
 
 
-def endeffector_position_objective(xyz, weight=1.0):
-  """
-  Create a position end effector objective with the given parameters.
-  Analogous to `EndEffectorPositionObjective
-  <http://docs.hebi.us/docs/cpp/cpp-1.0.0/classhebi_1_1robot__model_1_1
+def endeffector_position_objective(xyz: "VectorType", weight: float=1.0):
+  """Create a position end effector objective with the given parameters.
+  Analogous to `EndEffectorPositionObjective.
+
+  <https://files.hebi.us/docs/cpp/cpp-1.0.0/classhebi_1_1robot__model_1_1
   EndEffectorPositionObjective.html>`_ in the C++ API.
 
   :param xyz: list of x, y, and z position objective points
   :type xyz:  list, numpy.ndarray, ctypes.Array
-  
+
   :param weight: The weight of the objective
   :type weight:  int, float
 
   :return: the created objective
-  
+
   :raises ValueError: if xyz does not have at least 3 elements
   """
   xyz = _np.array(xyz, dtype=_np.float64, copy=True)
@@ -1313,21 +1362,21 @@ def endeffector_position_objective(xyz, weight=1.0):
   return Objective(xyz, weight)
 
 
-def endeffector_so3_objective(rotation, weight=1.0):
-  """
-  Create an SO3 end effector objective with the given parameters.
-  Analogous to `EndEffectorSO3Objective
-  <http://docs.hebi.us/docs/cpp/cpp-1.0.0/classhebi_1_1robot__model_1_1
+def endeffector_so3_objective(rotation: "VectorType", weight: float=1.0):
+  """Create an SO3 end effector objective with the given parameters. Analogous
+  to `EndEffectorSO3Objective.
+
+  <https://files.hebi.us/docs/cpp/cpp-1.0.0/classhebi_1_1robot__model_1_1
   EndEffectorSO3Objective.html>`_ in the C++ API.
 
   :param rotation: SO3 rotation matrix
   :type rotation:  str, list, numpy.ndarray, ctypes.Array
-  
+
   :param weight: The weight of the objective
   :type weight:  int, float
 
   :return: the created objective
-  
+
   :raises ValueError: if rotation matrix is not convertible to a 3x3 matrix,
                       or if the rotation matrix is not in the
                       `SO(3) <https://en.wikipedia.org/wiki/Rotation_group_SO(3)>`_
@@ -1366,21 +1415,21 @@ def endeffector_so3_objective(rotation, weight=1.0):
   return Objective(rotation, weight)
 
 
-def endeffector_tipaxis_objective(axis, weight=1.0):
-  """
-  Create a tip axis end effector objective with the given parameters.
-  Analogous to `EndEffectorTipAxisObjective
-  <http://docs.hebi.us/docs/cpp/cpp-1.0.0/classhebi_1_1robot__model_1_1
+def endeffector_tipaxis_objective(axis: "VectorType", weight: float=1.0):
+  """Create a tip axis end effector objective with the given parameters.
+  Analogous to `EndEffectorTipAxisObjective.
+
+  <https://files.hebi.us/docs/cpp/cpp-1.0.0/classhebi_1_1robot__model_1_1
   EndEffectorTipAxisObjective.html>`_ in the C++ API.
 
   :param axis: list of x, y, and z tipaxis objective points
   :type axis:  list, numpy.ndarray, ctypes.Array
-  
+
   :param weight: The weight of the objective
   :type weight:  int, float
 
   :return: the created objective
-  
+
   :raises ValueError: if axis does not have at least 3 elements
   """
   axis = _np.array(axis, dtype=_np.float64, copy=True)
@@ -1420,17 +1469,17 @@ def endeffector_tipaxis_objective(axis, weight=1.0):
   return Objective(axis, weight)
 
 
-def joint_limit_constraint(minimum, maximum, weight=1.0):
-  """
-  Create a joint limit constraint objective.
-  Analogous to `JointLimitConstraint
-  <http://docs.hebi.us/docs/cpp/cpp-1.0.0/classhebi_1_1robot__model_1_1
+def joint_limit_constraint(minimum: "VectorType", maximum: "VectorType", weight: float=1.0):
+  """Create a joint limit constraint objective. Analogous to
+  `JointLimitConstraint.
+
+  <https://files.hebi.us/docs/cpp/cpp-1.0.0/classhebi_1_1robot__model_1_1
   JointLimitConstraint.html>`_ in the C++ API.
 
-  :param minimum: 
+  :param minimum:
   :type minimum:  str, list, numpy.ndarray, ctypes.Array
 
-  :param maximum: 
+  :param maximum:
   :type maximum:  str, list, numpy.ndarray, ctypes.Array
 
   :param weight: The weight of the objective
@@ -1476,10 +1525,9 @@ def joint_limit_constraint(minimum, maximum, weight=1.0):
 
 
 def custom_objective(num_errors, func, user_data=None, weight=1.0):
-  """
-  Construct a custom objective using a provided function.
-  The `func` parameter is a function which accepts 3 parameters:
-  `positions`, `errors` and `user_data`.
+  """Construct a custom objective using a provided function. The `func`
+  parameter is a function which accepts 3 parameters: `positions`, `errors` and
+  `user_data`.
 
   The first two parameters are guaranteed to be numpy arrays
   with `dtype=np.float64`. The third parameter, `user_data`, may be `None`,
@@ -1514,9 +1562,9 @@ def custom_objective(num_errors, func, user_data=None, weight=1.0):
   class Objective(_ObjectiveBase):
 
     def __callback(self, user_data, num_positions, c_positions, c_errors):
-      """
-      The actual callback the C API invokes. Don't let the user mess
-      with ffi calls.
+      """The actual callback the C API invokes.
+
+      Don't let the user mess with ffi calls.
       """
       positions = _type_utils.np_array_from_dbl_ptr(c_positions, (num_positions,)).copy()
       errors = _type_utils.np_array_from_dbl_ptr(c_errors, (self._num_errors,))
